local options = MissionMinder.options;
local L = LibStub("AceLocale-3.0"):GetLocale("MissionMinder");
local DEBUG = MissionMinder.debug;

MissionMinder.planner = {};
local addon = MissionMinder.planner;



local function createInlineTexture(texture)
	return texture and ("|T" .. texture .. ":0|t") or nil;
end

local function canBeSentOnMission(follower)
	return follower and follower.isCollected and ( follower.state == nil
			or follower.state == GARRISON_FOLLOWER_ON_MISSION or follower.state == GARRISON_FOLLOWER_IN_PARTY);
end

local function isOnMission(follower)
	return follower and follower.status == GARRISON_FOLLOWER_ON_MISSION;
end

local function contains(list, value)
	for i = 1, #list do
		if(list[i] == value) then
			return true
		end
	end
	return false
end


local expGainTraits = { 80, 81, 29, 236, 98, 25, 99, 30 };
local function ignoreTrait(traitID, hasBonusXPReward)
	if ( options.IgnoreXPGainTraits() and not hasBonusXPReward and contains(expGainTraits, traitID)) then
		return true
	else
		return false
	end
end



local function getMissionRewards(mission)
	local rewards = C_Garrison.GetMissionRewardInfo(mission.missionID);

	local t = {};
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			local xp = reward.followerXP;
			if ( xp >= 1000 ) then
				xp = xp/1000 .. L["K"];
			end
			tinsert(t, "+" .. xp .. " " .. createInlineTexture(reward.icon));
		elseif (reward.itemID) then
			local itemTexture = select(10, GetItemInfo(reward.itemID));
			if(reward.quantity == 1) then
				tinsert(t, createInlineTexture(itemTexture));
			else
				tinsert(t, reward.quantity .. "x " .. createInlineTexture(itemTexture));
			end
		elseif(reward.currencyID) then
			if (reward.currencyID == 0) then -- gold
				tinsert(t, GetCoinTextureString(reward.quantity));
			else
				tinsert(t, reward.quantity .. " " .. createInlineTexture(reward.icon));
			end
		else
			-- because this is tooltip based, we ignore the value instead of writing the name
		end
	end

	return table.concat(t, ", ");
end

local function MissionHasBonusXPReward(missionID)
	local rewards = C_Garrison.GetMissionRewardInfo(missionID);
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			return true;
		end
	end

	return false;
end

local timeLeft = {};
local inProgressMissions = {};
local function updateFollowerMissionTimeLeft(timeLeft)
	C_Garrison.GetInProgressMissions(inProgressMissions);

	if (inProgressMissions and #inProgressMissions > 0) then
		for _,mission in pairs(inProgressMissions) do
			for _,id in pairs(mission.followers) do
				-- Old values are overwritten, but if a follower is AVAILABLE any previous values will still be present.
				--Thats Okay because we only use this data if the follower is GARRISON_FOLLOWER_ON_MISSION.
				timeLeft[id] = {
					["missionID"] = mission.missionID,
					["remaining"] = mission.timeLeft;
				};
			end
		end
	end
end

local function getFollowerMissionBiasRGB(missionID,followerID)
	local bias = C_Garrison.GetFollowerBiasForMission(missionID, followerID);

	if ( bias == -1 ) then -- Impossible
		return 1, 0.1, 0.1;
	elseif ( bias < 0 ) then -- Hard
		return 1, 0.5, 0.25;
	else -- Normal
		return 1, 1, 1;
	end
end

local function levelComparator(left, right)
	if (left.iLevel ~= right.iLevel ) then
		return left.iLevel > right .iLevel;
	else
		return left.level > right.level;
	end
end

local function getMissionTimeleft(follower)
	if isOnMission(follower) then
		-- entry can be nil if follower is on a complete mission that has been viewed, but still needs to be looted.
		local entry = timeLeft[follower.followerID];
		if ( entry == nil or MissionMinder.IsMissionComplete(entry) ) then
			return L["completed"], nil, nil, nil; -- Default yellow
		else
			return string.format("(%s)", timeLeft[follower.followerID].remaining), 0.7, 0.7, 0.7; -- Gray
		end
	elseif (follower.status == GARRISON_FOLLOWER_WORKING) then
		return L["working"], 1, 0, 0; -- red
	else
		return L["available"], 0, 1, 0; -- Green
	end
end



local cache = {
	["isDirty"] = true,
	["abilities"] = {},
	["traits"] = {}
};





local function dirtyMissionCounterCache()
	cache.isDirty = true;
end

local function getFollowerCountersForMisison(followerID, missionID)
	if( cache.isDirty or not cache.abilities[missionID] or not cache.traits[missionID] ) then
		cache.abilities[missionID] = C_Garrison.GetBuffedFollowersForMission(missionID);
		cache.traits[missionID] = C_Garrison.GetFollowersTraitsForMission(missionID);
	end

	local abilities = cache.abilities[missionID];
	local traits = cache.traits[missionID];

	local hasBonusXP = MissionHasBonusXPReward(missionID);

	local counters = {};
	local hasNonFilterable = false
	if ( abilities and abilities[followerID] ) then
		for i = 1, #abilities[followerID] do
			local ability = abilities[followerID][i];
			counters[#counters + 1] = createInlineTexture(ability.icon);

			hasNonFilterable = true
		end
	end

	if ( traits and traits[followerID] ) then
		for i = 1, #traits[followerID] do
			local trait = traits[followerID][i];
			counters[#counters + 1] = createInlineTexture(trait.icon);

			if ( not ignoreTrait(trait.traitID, hasBonusXP) ) then
				hasNonFilterable = true
			end
		end
	end

	if ( hasNonFilterable ) then
		return counters;
	else
		return nil;
	end
end

local function createMissionTooltip(self, button)
	local mission = self.info;

	if(mission.inProgress) then -- Do not show tooltip on inProgress missions
		return;
	end

	updateFollowerMissionTimeLeft(timeLeft);

	local followers = C_Garrison.GetFollowers(); -- GarrisonMissionFrame.FollowerList.followers
	local numFollowers = followers and #followers or 0;
	table.sort(followers, levelComparator);

	local counterAvailable = false;
	for i1 = 1, numFollowers do -- Necessary to walk all followers because we want to sort by level
		local follower = followers[i1];

		if ( canBeSentOnMission(follower) ) then
			local r1, g1, b1 = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);

			local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);
			if ( counters and #counters > 0 ) then
				if ( not counterAvailable ) then
					GameTooltip:AddLine(" ");
					GameTooltip:AddLine(L["Counters"]);
					counterAvailable = true;
				end

				local level = follower.level
				if (mission.iLevel and mission.iLevel > 600 and follower.level == GARRISON_FOLLOWER_MAX_LEVEL) then
					level = follower.iLevel
				end

				local left = string.format("[%d] %s %s", level, follower.name, table.concat(counters, ""));
				local right, r2, g2, b2 = getMissionTimeleft(follower);
				GameTooltip:AddDoubleLine(left, right, r1, g1, b1, r2, g2, b2);
			end
		end
	end

	if ( not counterAvailable ) then
		GameTooltip:AddLine(" ");
		GameTooltip:AddLine(L["No counters available"]);
	end

	GameTooltip:Show(); -- resize the tooltip for the new lines
end

local function createFollowersTooltip(self)
	local follower = C_Garrison.GetFollowerInfo(self.id);
	if ( not follower.isCollected ) then
		return;
	end

	GameTooltip:SetOwner(self, "ANCHOR_NONE");
	GameTooltip:SetPoint("TOPRIGHT", self, "TOPLEFT", -10);

	local r, g, b = GetItemQualityColor(follower.quality);
	GameTooltip:SetText(follower.name, r, g, b);
	GameTooltip:AddLine(" ");

	local availableMissions = GarrisonMissionFrame.MissionTab.MissionList.availableMissions;
	C_Garrison.GetAvailableMissions(availableMissions);
	table.sort(availableMissions, levelComparator);

	local atleastOneMission = false;
	for _,mission in pairs(availableMissions) do
		local r, g, b = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);

		local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);
		if ( counters and #counters > 0 ) then
			atleastOneMission = true;

			local level = mission.level
			if (mission.iLevel and mission.iLevel > 600) then
				level = mission.iLevel
			end

			local rewards = getMissionRewards(mission);
			local left = string.format("[%d] %s (%s)", level, mission.name, rewards);
			local right = table.concat(counters, "");
			GameTooltip:AddDoubleLine(left, right, r, g, b, 1, 1, 1);
		end
	end

	if ( not atleastOneMission ) then
		GameTooltip:AddLine(L["No counters available"]);
	end

	GameTooltip:Show();
	cache.isDirty = false;
end

local function biasColorFollowerLevels()
	local mission = GarrisonMissionFrame.MissionTab.MissionPage.missionInfo;
	local buttons = GarrisonMissionFrame.FollowerList.listScroll.buttons;

	for i = 1, #buttons do
		local button = buttons[i];
		local r, g, b = 1, 1, 1;

		if ( GarrisonMissionFrame.MissionTab.MissionPage:IsShown() and not GarrisonMissionFrame.FollowerTab:IsShown()
				and mission and button.isCollected) then
			r, g, b = getFollowerMissionBiasRGB(mission.missionID, button.id);
		end

		button.PortraitFrame.Level:SetTextColor(r, g, b)
	end
end

local function hideMissionCountersTooltip(self)
	GameTooltip:Hide();
end


local singleclickTimer = nil;
-- Override the standard OnClick event to delay just long enough to determine if it is a DoubleClick.
--If it is a double click then the OnClick event will not fire. If it isn't a double click then the
--OnClick event (which will run via timer) will still run; albeit slightly delayed.
local function followerSingleClick(self, button)
	if(singleclickTimer) then
		singleclickTimer:Cancel();
	end

	-- Only delay OnClick in very specific cases
	if ( self.isCollected and button == "LeftButton" ) then
		if ( GarrisonMissionFrame.MissionTab.MissionPage:IsShown() and not GarrisonMissionFrame.FollowerTab:IsShown() ) then
			singleclickTimer = C_Timer.NewTimer(0.2, function()
				addon.GarrisonFollowerListButton_OnClick(self, button);
			end);
			return;
		end
	end

	addon.GarrisonFollowerListButton_OnClick(self, button);
end

-- Handle OnDoubleClick event and disable the single click timer
local function assignFollowerToMission(self, button)
	if(singleclickTimer) then
		singleclickTimer:Cancel();
	end

	-- Hide the rightclick menu
	GarrisonFollowerOptionDropDown.followerID = nil;
	CloseDropDownMenus();

	if ( self.isCollected and self.info.status == nil and button == "LeftButton" ) then
		local mission = GarrisonMissionFrame.MissionTab.MissionPage.missionInfo;
		if ( mission and GarrisonMissionFrame.MissionTab.MissionPage:IsShown()
					and not GarrisonMissionFrame.FollowerTab:IsShown() ) then
			DEBUG("assignFollowerToMission(mission=%s, follower=%s)", mission.missionID, self.id);
			GarrisonMissionPage_AddFollower(self.id);
		end
	end
end

local function disableTimer()
	if(singleclickTimer) then
		singleclickTimer:Cancel()
		singleclickTimer = nil
	end
end

function addon.initialize()
	local missionButtons = GarrisonMissionFrame.MissionTab.MissionList.listScroll.buttons;
	for i = 1, #missionButtons do
		missionButtons[i]:HookScript("OnEnter", createMissionTooltip);
	end

	addon.GarrisonFollowerListButton_OnClick = GarrisonFollowerListButton_OnClick;

	local followerButtons = GarrisonMissionFrame.FollowerList.listScroll.buttons;
	for i = 1, #followerButtons do
		followerButtons[i]:SetScript("OnEnter", createFollowersTooltip);
		followerButtons[i]:SetScript("OnLeave", hideMissionCountersTooltip);

		-- Allow double click to assign followers to mission
		followerButtons[i]:SetScript("OnClick", followerSingleClick);
		followerButtons[i]:SetScript("OnDoubleClick", assignFollowerToMission);
	end

	-- Color follower levels by bias
	hooksecurefunc("GarrisonFollowerList_Update", biasColorFollowerLevels)

	-- Make sure we update the list after showing the mission page. This is important because
	--the mission info is not initially available when the list first shows
	hooksecurefunc("GarrisonMissionPage_ShowMission", biasColorFollowerLevels);

	-- Disable time on MissionPage:OnHide
	GarrisonMissionFrame.MissionTab.MissionPage:HookScript("OnHide", disableTimer);
end


local events = MissionMinder.events;
events:RegisterEvent("GARRISON_FOLLOWER_LIST_UPDATE", dirtyMissionCounterCache);
events:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED", dirtyMissionCounterCache);
events:RegisterEvent("GARRISON_FOLLOWER_REMOVED", dirtyMissionCounterCache);